#include <iostream>
#include <cmath>
#include <cstdio>
#include <utility>
#include "bwm_video_corr_processor.h"
//:
// \file
#include <bwm/video/bwm_video_site_io.h>
#include <bwm/video/bwm_video_corr.h>
#include <vnl/vnl_numeric_traits.h>
#include <vnl/vnl_double_3.h>
#include <vnl/algo/vnl_levenberg_marquardt.h>
#include <vnl/algo/vnl_amoeba.h>
#include <vul/vul_file.h>
#include <vpl/vpl.h> // vpl_unlink()
#include <bpgl/algo/bpgl_interpolate.h>
#include <vpgl/algo/vpgl_bundle_adjust.h>
#include <vgl/vgl_point_2d.h>
#include <vgl/vgl_homg_point_3d.h>
#include <vgl/vgl_distance.h>
#include <vgl/algo/vgl_rotation_3d.h>

#include <bsta/bsta_histogram.h>
#include <brip/brip_vil_float_ops.h>
#include <bwm/video/bwm_video_cam_istream.h>
#include <bwm/video/bwm_video_cam_ostream.h>
#include <vidl/vidl_frame.h>
#include <vidl/vidl_convert.h>
#include <vidl/vidl_image_list_istream.h>

#include <vpgl/vpgl_lvcs.h>

#ifdef _MSC_VER
#  include <vcl_msvc_warnings.h>
#endif

//Minimum number of correspondences on a frame to compute cameras
//static const unsigned min_corrs = 10;
//static const unsigned min_corrs = 9;
static const unsigned min_corrs = 6;
// if an element of pixels is negative it indicates the point was outside the
// image
static void extract_window(float u, float v, int radius,
                           const vil_image_view<float>& image,
                           std::vector<float>& pixels)
{
  pixels.clear();
  float umin = u-radius, umax = u+radius;
  float vmin = v-radius, vmax = v+radius;
  int ni = image.ni(), nj = image.nj();
  for (float vi = vmin; vi<=vmax; ++vi)
    for (float ui = umin; ui<=umax; ++ui)
      if (ui<0||vi<0||ui>=ni||vi>=nj)
        pixels.push_back(-1.0f);
      else {
        double val = brip_vil_float_ops::bilinear_interpolation(image, ui, vi);
        pixels.push_back(static_cast<float>(val));
      }
}

bwm_video_corr_lsqr_cost_func::
bwm_video_corr_lsqr_cost_func(vil_image_view<float> const& base_image,
                              unsigned match_radius,
                              std::vector<float> corr_window_ab
                             )
  : vnl_least_squares_function(1,1), base_image_(base_image),
    match_radius_(match_radius), corr_window_ab_(std::move(corr_window_ab))
{
  use_gradient_ = false;
  // a mxm image region centered on the projected world pointa
  unsigned m = (2*match_radius_+1);
  vnl_least_squares_function::init(2, m*m);
}

//the least squares cost function for minimizing intensity differences
// x is a 2 element vector for correspondence position
// f is an M-element vector, where M = (2*match_radius_+1)^2
// i.e. a window around each projected world point.
void bwm_video_corr_lsqr_cost_func::f(vnl_vector<double> const& x,
                                      vnl_vector<double>& fx)
{
  auto u = static_cast<float>(x[0]);
  auto v = static_cast<float>(x[1]);
  std::vector<float> pixels;
  extract_window(u, v, match_radius_, base_image_, pixels);
  unsigned msq = pixels.size();
  for (unsigned k = 0; k<msq; ++k)
    if (corr_window_ab_[k]<0||pixels[k]<0)
      fx[k] = 0.0;
    else
      fx[k] = corr_window_ab_[k] - pixels[k];
#if 0
  std::cout << "\nX = (" << x[0] << ' ' << x[1] << '\n'
           << "f\n";
  std::cout.precision(3);
  for (unsigned i = 0; i< fx.size(); ++i)
    std::cout<< fx[i] << ' ';
  std::cout << '\n';
#endif
}

bwm_video_corr_cost_function::
bwm_video_corr_cost_function(vil_image_view<float> const& base_image,
                             unsigned match_radius,
                             std::vector<float> corr_window_ab)
  : vnl_cost_function(2), base_image_(base_image),
    match_radius_(match_radius), corr_window_ab_(std::move(corr_window_ab))
{}

double bwm_video_corr_cost_function::f(vnl_vector<double> const& x)
{
  auto u = static_cast<float>(x[0]);
  auto v = static_cast<float>(x[1]);
  double r = 0;
  std::vector<float> pixels;
  extract_window(u, v, match_radius_, base_image_, pixels);
  unsigned msq = pixels.size();
  for (unsigned k = 0; k<msq; ++k)
    if (corr_window_ab_[k]<0||pixels[k]<0)
      continue;
    else {
      double res = corr_window_ab_[k] - pixels[k];
      r += res*res;
    }
  return r;
}

bool bwm_video_corr_processor::open_video_stream(std::string const& video_path)
{
  if (video_path=="")
    return false;
  //for now assume we are opening an image_list codec
  video_istr_ = new vidl_image_list_istream(video_path);
  bool open = video_istr_->is_open();
  if (open)
    video_istr_->seek_frame(0);
  return open;
}

bool bwm_video_corr_processor::
open_camera_istream(std::string const& camera_path)
{
  if (camera_path=="")
    return false;
  cam_istr_ = new bwm_video_cam_istream(camera_path);
  bool open = cam_istr_->is_open();
  if (open) {
    std::cout << "in bwm_video_corr_processor::open_camera_istream() -- input stream: " << camera_path << " is opened!\n";
    cam_istr_->seek_camera(0);
  }
  else {
    std::cout << "in bwm_video_corr_processor::open_camera_istream() -- input stream: " << camera_path << " can NOT be opened!\n";
  }
  return open;
}

bool bwm_video_corr_processor::
open_camera_ostream(std::string const& camera_path)
{
  if (camera_path=="")
    return false;
  cam_ostr_ = new bwm_video_cam_ostream(camera_path);
  bool open = cam_ostr_->is_open();
  if (open) {
    std::cout << "in bwm_video_corr_processor::open_camera_ostream() -- output stream: " << camera_path << " is opened!\n";
  }
  else {
    std::cout << "in bwm_video_corr_processor::open_camera_ostream() -- output stream: " << camera_path << " can NOT be opened!\n";
  }
  return open;
}

void bwm_video_corr_processor::close_camera_ostream()
{
  cam_ostr_->close();
}

void bwm_video_corr_processor::close_camera_istream()
{
  cam_istr_->close();
}


void remove_dir(std::string const& dir)
{
  std::string glob = dir + "/*";
  vul_file::delete_file_glob(glob.c_str());
  vpl_unlink(dir.c_str());
}

bool bwm_video_corr_processor::open_video_site(std::string const& site_path,
                                               bool cameras_exist)
{
  if (!site_io_.open(site_path))
    return false;
  site_path_ = site_path;
  site_name_ = site_io_.name();
  video_path_ = site_io_.video_path();
  camera_path_ = site_io_.camera_path();
  if (!this->open_video_stream(video_path_))
    return false;
  if (cameras_exist){
    if (!this->open_camera_istream(camera_path_))
      return false;
  }
  else {
    std::string dir = vul_file::dirname(site_path)+"/cameras";
    if (vul_file::exists(dir))
      remove_dir(dir);
    if (!vul_file::make_directory_path(dir))
      return false;
    camera_path_ = dir+"/*";
    if (!this->open_camera_ostream(dir))
      return false;
  }
  std::vector<bwm_video_corr_sptr> corrs = site_io_.corrs();
  if (!corrs.size())
    return false;
  this->set_correspondences(corrs);
  return true;
}

void bwm_video_corr_processor::
set_world_pts(std::vector<vgl_point_3d<double> > const& pts)
{
  auto pit = pts.begin();
  for (auto cit = corrs_.begin();
       cit != corrs_.end()&&pit!=pts.end(); ++cit, ++pit)
    (*cit)->set_world_pt(*pit);
}

std::vector<vgl_point_3d<double> > bwm_video_corr_processor::world_pts()
{
  std::vector<vgl_point_3d<double> > pts;
  for (auto & corr : corrs_)
    if (corr->world_pt_valid())
      pts.push_back(corr->world_pt());
  return pts;
}

//: if the world coordinates are given in global coordinates of satellite cameras, convert them to local coordinate frame of the given lvcs
void bwm_video_corr_processor::convert_world_pts_to_local(const vpgl_lvcs_sptr& lvcs)
{
  for (auto & corr : corrs_)
  {
    if (corr->world_pt_valid()) {
      vgl_point_3d<double> pt = corr->world_pt();
      double x,y,z;
      lvcs->global_to_local(pt.x(), pt.y(), pt.z(), lvcs->get_cs_name(), x,y,z);
      vgl_point_3d<double> new_pt(x,y,z);
      std::cout << "world pt: " << pt << " converted to " << new_pt << std::endl;
      corr->set_world_pt(new_pt);
    }
  }
}

bool bwm_video_corr_processor::write_video_site(std::string const& site_path)
{
  if (site_name_ == "")
    return false;
  bwm_video_site_io sio;
  sio.set_name(site_name_);
  sio.set_video_path(video_path_);
  sio.set_camera_path(camera_path_);
  sio.set_corrs(corrs_);
  sio.x_write(site_path);
  return true;
}

void bwm_video_corr_processor::write_cameras_txt(std::string const& cam_txt_dir, std::vector<vpgl_perspective_camera<double> > const& cameras)
{
  char filename[1024];
  if (vul_file::is_directory(cam_txt_dir.c_str()))
  {
    std::cout << "writing " << cameras.size() << " cams to: " << cam_txt_dir << std::endl;
    for (unsigned i=0;i<cameras.size();i++)
    {
      std::sprintf(filename,"%s/camera%05d.txt",cam_txt_dir.c_str(),i);
      std::ofstream ofile(filename);
      if (ofile)
      {
        ofile<<cameras[i].get_calibration().get_matrix()<<'\n'
             <<cameras[i].get_rotation().as_matrix()<<'\n'
             <<cameras[i].get_translation().x()<<' '<<cameras[i].get_translation().y()<<' '<<cameras[i].get_translation().z()<<'\n';
      }
    }
  }
}

bool bwm_video_corr_processor::
interpolate_cameras(std::vector<vpgl_perspective_camera<double> > known_cams,
                    std::vector<bool> unknown,
                    std::vector<vpgl_perspective_camera<double> >& interp_cams)
{
  unsigned ncameras = unknown.size();
  interp_cams.resize(ncameras);

  //fill in the known cameras
  unsigned ne = unknown.size();
  for (unsigned fi=0, count=0; fi<ne; ++fi)
    if (!unknown[fi])
      interp_cams[fi]=known_cams[count++];
  //then interpolate the missing cameras
  for (unsigned fi = 0; fi<ne; ++fi)
  {
    if (unknown[fi])//row is empty
    {
      // case 1 - i == 0
      if (fi==0)
      {
        // find the smallest row that is not empty
        bool found = false;
        for (unsigned j = 1; j<ne&&!found; ++j)
          if (!unknown[j])
          {
            found = true;
            for (unsigned k = 0; k<j; ++k)
              interp_cams[k] = known_cams[0];
            fi = j;
          }
        if (!found){
          std::cerr << "Shouldn't happen\n";
          return false;}
      }
      // case 2 - i == ne-1 (the last row)
      if (fi == ne-1)//last known camera
        interp_cams[fi] = known_cams[known_cams.size()-1];

      // case 3 - the normal case
      // typically there will be a run of missing cameras
      // also the cameras are filled in up to fi-1;
      // find the next known camera
      bool found = false;
      unsigned ki = fi+1;
      for (; ki<ne&&!found; ++ki)
        if (!unknown[ki])
          found = true;
      if (!found){
        std::cerr << "Shouldn't happen\n";
        return false;}
      // get the cameras at fi-1 and next solved camera
      vpgl_perspective_camera<double> c0 = interp_cams[fi-1];
      vpgl_perspective_camera<double> c1 = interp_cams[ki-1];
      std::vector<vpgl_perspective_camera<double> > icams;
      unsigned n_interp = ki-fi-1;
      if (verbose_)
        std::cout << "Interpolating between frames[" << fi-1
                 << ':' << ki-1 << "]\n";
      if (!bpgl_interpolate::interpolate(c0, c1, n_interp, icams))
      {
        std::cerr << "Interpolation failed\n";
        return false;
      }
      for (unsigned c = 0; c<icams.size(); ++c)
        interp_cams[fi+c] = icams[c];
      fi = ki-1;
    }
  }
  if (verbose_){
    std::cout << "Solved and Interpolated Cameras\n";
    for (unsigned c = 0; c<ncameras; ++c){
      vgl_point_3d<double> p =interp_cams[c].get_camera_center();
      vgl_rotation_3d<double> R = interp_cams[c].get_rotation();
      std::cout << "C[" << c << "]("<< R << '|' << p.x() << ' ' << p.y() << ' '
               << p.z() <<  ")\n"; }
  }
  return true;
}

void min_max_frame(std::vector<bwm_video_corr_sptr> const& corrs,
                   unsigned& min_frame, unsigned& max_frame)
{
  min_frame = vnl_numeric_traits<unsigned>::maxval;
  max_frame = 0;
  for (const auto & corr : corrs)
    {
      unsigned minf = corr->min_frame();
      unsigned maxf = corr->max_frame();
      if (minf<min_frame)
        min_frame = minf;
      if (maxf>max_frame)
        max_frame = maxf;
    }
}

//: return the number of cameras that observe the correspondences
unsigned bwm_video_corr_processor::get_ncameras(unsigned& min_frame, unsigned& max_frame) const
{
  min_frame =0; max_frame = 0;
  //get the start and end frame numbers. assume contiguous frames inbetween
  min_max_frame(corrs_, min_frame, max_frame);
  if (verbose_)
    std::cout << "Start frame = " << min_frame << " End frame = " << max_frame << '\n';
  return max_frame-min_frame +1;
}

//: compute the mask on existing correspondences
// The mask is m x n , where m = number of cameras and
// n = number of world points (corrs)
void bwm_video_corr_processor::mask(unsigned& min_frame, unsigned& max_frame,
                                    std::vector<std::vector<bool> >& mask)
{
  unsigned ncameras = get_ncameras(min_frame, max_frame);
  unsigned npoints = corrs_.size();
  std::vector<bool> init(npoints);
  mask.resize(ncameras, init);

  for (unsigned w = 0; w<npoints; ++w){
    unsigned f_rel = 0;//relative frame number
    for (unsigned f = min_frame; f<=max_frame; ++f, ++f_rel)
    {
      vgl_point_2d<double> pt;
      mask[f_rel][w] = corrs_[w]->match(f, pt);
    }
  }
}

bool bwm_video_corr_processor::
initialize_world_pts_and_cameras(vpgl_calibration_matrix<double> const& K,
                                 double initial_depth)
{
  unsigned min_frame =0, max_frame = 0;
  std::vector<std::vector<bool> > mask;
  this->mask(min_frame, max_frame, mask);
  //The implementation is not general enough yet to handle
  //a non-zero start frame
  if (min_frame)
    return false;
  unsigned ncameras = max_frame-min_frame +1;
  unsigned npoints = corrs_.size();
  if (verbose_)
    std::cout << "Initializing " << ncameras << " cameras on " << npoints
             << " correspondences\n";
  std::vector<vgl_point_2d<double> > image_points(ncameras*npoints);
  for (unsigned w = 0; w<npoints; ++w)
  {
    unsigned f_rel = 0;//relative frame num == 0 at min_frame
    for (unsigned f = min_frame; f<=max_frame; ++f, ++f_rel)
    {
      vgl_point_2d<double> pt;
      if (corrs_[w]->match(f, pt))
        image_points[f_rel*npoints + w] = pt;
    }
  }
  //find the number of rows that are completely masked and
  //remove them creating a smaller problem
  std::vector<bool> unknown_frames(ncameras, false);
  int filled_rows = 0;
  for (unsigned f_rel = 0; f_rel<ncameras; ++f_rel){
    bool empty = true;
    for (unsigned w = 0; w<npoints; ++w)
      if (mask[f_rel][w]) empty = false;
    if (empty)
      unknown_frames[f_rel] = true;
    else
      ++filled_rows;
  }
  if (verbose_)
    std::cout << "Executing bundle adjustment on " << filled_rows
             << " cameras\n";
  //create a new mask array and image point vector for the smaller problem
  std::vector<bool> init(npoints);
  std::vector<std::vector<bool> > cmask(filled_rows, init);
  std::vector<vgl_point_2d<double> > cimage_points;
  unsigned ff = 0;
  for (unsigned f = 0; f<ncameras; ++f)
    if (!unknown_frames[f])
    {
      unsigned cnt = 0;
      for (unsigned w = 0; w<npoints; ++w){
        cmask[ff][w] = mask[f][w];
        if (mask[f][w]){ cnt++;
        cimage_points.push_back(image_points[f*npoints + w]);}
      }
      if (cnt<min_corrs){
        std::cout << "NCorrs[" << f << "]= " << cnt << '\n';
        for (unsigned w = 0; w<npoints; ++w)
          if (mask[f][w])
          {
            vgl_point_2d<double> pt;
            if (corrs_[w]->match(f, pt))
              std::cout << "bad corr " << pt << '\n';
          }
      }
      ++ff;
    }
#if 0
  //sanity check on image points
  unsigned kk = 0;
  for (unsigned ff = 0; ff<filled_rows; ++ff)
    for (unsigned w = 0; w<npoints; ++w)
      if (cmask[ff][w]){
        std::cout << "Image Point[" << ff << "][" << w
                 << "]=" << cimage_points[kk] << '\n';
        kk++;
      }
#endif
  vnl_double_3 r(0,0,0.000001);
  vgl_rotation_3d<double> I(r); // small initial rotation
  vgl_homg_point_3d<double> center(0.0, 0.0, -initial_depth);
  vpgl_perspective_camera<double> default_camera(K, center, I);

  std::vector<vpgl_perspective_camera<double> > unknown_cameras(filled_rows,
                                                               default_camera);
  // initialize unknown world points
  vgl_point_3d<double> pun(0.0, 0.0, 0.0);
  std::vector<vgl_point_3d<double> > unknown_world(npoints,pun);

  // execute the bundle adjustment
  vpgl_bundle_adjust adj;
  /*bool success=*/ adj.optimize(unknown_cameras,
                                 unknown_world, cimage_points,
                                 cmask);
  //save the solved world points in the correspondences
  for (unsigned w = 0; w<npoints; ++w)
  {
    vgl_point_3d<double> pt = unknown_world[w];
    corrs_[w]->set_world_pt(pt);
  }
  //WARNING!! THIS FUNCTION NOT VALID FOR MIN_FRAME != 0
  if (!this->interpolate_cameras(unknown_cameras, unknown_frames, cameras_))
    return false;
  return true;
}

bool bwm_video_corr_processor::write_cameras_to_stream()
{
  if (cam_ostr_&&cam_ostr_->is_open()){
    for (auto & camera : cameras_)
      cam_ostr_->write_camera(&camera);
    return true;
  }
  return false;
}

bool bwm_video_corr_processor::frame_at_index(unsigned frame_index,
                                              vil_image_view<float>& view)
{
  // extract the image from the frame
  if (!video_istr_) return false;
  if (!video_istr_->seek_frame(frame_index)) return false;
  vidl_frame_sptr frame = video_istr_->current_frame();
  if (!frame)
  {
    std::cerr << "Failed to seek to frame " << frame_index << '\n';
    return false;
  }
  else if (frame->pixel_format() == VIDL_PIXEL_FORMAT_MONO_16){
    static vil_image_view<vxl_uint_16> img;
    if (vidl_convert_to_view(*frame,img))
      view = brip_vil_float_ops::convert_to_float(img);
    else{
      std::cerr << "Failed to convert frame to vil_image_view\n";
      return false;
    }
  }
  else{
    static vil_image_view<vxl_byte> img;
    if (vidl_convert_to_view(*frame,img,VIDL_PIXEL_COLOR_RGB)) {
      view = brip_vil_float_ops::convert_to_float(img);
    }
    else{
      std::cerr << "Failed to convert frame to vil_image_view\n";
      return false;
    }
  }
  return true;
}

// compute the set of image samples around each correspondence
// at the start frame and end frame of a tracking interval (a, b).
void bwm_video_corr_processor::
compute_ab_corr_windows(unsigned match_radius,
                        std::vector<bool> const& mask_a,
                        std::vector<bool> const& mask_b)
{
  //n is the number of correspondences to find, i.e. determine (u, v) for each.
  unsigned n = corrs_.size();
  vgl_point_2d<double> ini(-1, -1);
  std::vector<vgl_point_2d<double> > corrs_a(n, ini), corrs_b(n, ini);
  for (unsigned i = 0; i<n; ++i){
    if (!(mask_a[i]&&mask_b[i])) continue;
    bwm_video_corr_sptr c = corrs_[i];
    vgl_point_2d<double> ipt;
    if (c->match(frame_index_a_, ipt)) corrs_a[i]=ipt;
    else continue;
    if (c->match(frame_index_b_, ipt)) corrs_b[i]=ipt;
    else continue;
  }
  //extract the correspondence image windows
  //bilinear interpolation is used to sample the input images
  for (unsigned i = 0; i<n; ++i)
  {
    if (!(mask_a[i]&&mask_b[i])) continue;
    vgl_point_2d<double> ipta = corrs_a[i], iptb = corrs_b[i];
    if (ipta.x()<0||ipta.y()<0||iptb.x()<0||iptb.y()<0)
      continue;
    auto ua = static_cast<float>(ipta.x());
    auto va = static_cast<float>(ipta.y());
    std::vector<float> pixels_a;
    extract_window(ua, va, match_radius, image_a_, pixels_a);
    auto ub = static_cast<float>(iptb.x());
    auto vb = static_cast<float>(iptb.y());
    std::vector<float> pixels_b;
    extract_window(ub, vb, match_radius, image_b_, pixels_b);
    corr_windows_a_.push_back(pixels_a);
    corr_windows_b_.push_back(pixels_b);
  }
}

void bwm_video_corr_processor::
set_correspondences(std::vector<bwm_video_corr_sptr> const& corrs)
{
  corrs_ = corrs;
}

void bwm_video_corr_processor::
exhaustive_init(vnl_vector<double>& position,
                unsigned win_radius,
                unsigned search_radius,
                vil_image_view<float> const& base,
                std::vector<float> corr_win,
                double& start_error,
                double& end_error)
{
  auto u0 = static_cast<float>(position[0]);
  auto v0 = static_cast<float>(position[1]);
  float uf = 0, vf = 0;
  float umin = u0-search_radius, umax = u0+search_radius;
  float vmin = v0-search_radius, vmax = v0+search_radius;
  if (umin < 0) umin = 0;
  auto ni = static_cast<float>(base.ni()), nj = static_cast<float>(base.nj());
  if (umax>=ni) umax = ni-1;
  if (vmin < 0) vmin = 0;
  if (vmax>=nj) vmax = nj-1;
  unsigned pix_cnt = 0;
  double rmin = vnl_numeric_traits<double>::maxval;
  for (float v = vmin; v<=vmax; ++v)
    for (float u = umin; u<=umax; ++u)
    {
      std::vector<float> pixels;
      extract_window(u, v, win_radius, base, pixels);
      unsigned msq = pixels.size();
      double r = 0;
      for (unsigned k = 0; k<msq; ++k)
        if (corr_win[k]<0||pixels[k]<0)
          continue;
        else {
          double res = corr_win[k] - pixels[k];
          r += res*res;
          pix_cnt++;
        }
      if (pix_cnt)
        r/=pix_cnt;
      else
        return;
      if (u == u0&&v == v0)
        start_error = std::sqrt(r);
      if (r<rmin){
        rmin = r;
        uf = u; vf = v;
      }
    }
  end_error = std::sqrt(rmin);
  if (uf == umin || uf == umax || vf == vmin || vf == vmax)
    end_error = -1.0;

  position[0]=uf;   position[1]=vf;
}

// it is assumed that there is a correct correspondence at frame a,
// and at frame b. The approach is to minimize the sum of squared
// differences between the image intensities of the known correspondences
// and the missing correspondence at each frame between the known frames.
// The search starts at the projected world point location in each
// between frame
bool bwm_video_corr_processor::find_missing_corrs(unsigned frame_index_a,
                                                  std::vector<bool> mask_a,
                                                  unsigned frame_index_b,
                                                  std::vector<bool> mask_b,
                                                  unsigned frame_index_x,
                                                  unsigned win_radius,
                                                  unsigned search_radius,
                                                  bool  /*use_lmq*/)
{
  unsigned n = corrs_.size();
  if (!n) return false;
  //all correspondences must have a world point for the
  //algorithm to work
  if (!world_pts_valid_){
    world_pts_valid_ = true;
    for (unsigned i = 0; i<n&&world_pts_valid_; ++i)
    {
      bwm_video_corr_sptr c = corrs_[i];
      if (c->world_pt_valid())
        world_pts_.push_back(c->world_pt());
      else
        world_pts_valid_ = false;
    }
  }
  if (!world_pts_valid_){
    std::cerr << " world points not valid\n";
      return false;
  }
  //check for index bounds consistency
  if (frame_index_a>=frame_index_b ||
      frame_index_x <= frame_index_a ||
      frame_index_x >= frame_index_b)
    return false;
  //bounding frames may already be cached
  bool compute_ab = false;
  if (frame_index_a != frame_index_a_)
  {
    if (!this->frame_at_index(frame_index_a, image_a_))
      return false;
    frame_index_a_ = frame_index_a;
    compute_ab = true;
  }
  if (frame_index_b != frame_index_b_)
  {
    if (!this->frame_at_index(frame_index_b, image_b_))
      return false;
    frame_index_b_ = frame_index_b;
    compute_ab = true;
  }
  if (verbose_)
    std::cout << "Finding correspondences in frame interval ["
             << frame_index_a << ':' << frame_index_b << "]\n";
  if (compute_ab)
    this->compute_ab_corr_windows(win_radius, mask_a, mask_b);
  vil_image_view<float> fvx;
  if (!this->frame_at_index(frame_index_x, fvx))
    return false;
  //get the camera at frame x
  if (!cam_istr_) {
    std::cerr << "No camera stream\n";
    return false;
  }
  if (!cam_istr_->seek_camera(frame_index_x)){
    std::cerr << "Can't seek to camera frame " << frame_index_x << '\n';
    return false;
  }
  vpgl_perspective_camera<double>* camx = cam_istr_->current_camera();

  double nres = 2*win_radius +1;
  nres *= nres;
  //to find the nearest frame to the search frame
  unsigned da = frame_index_x - frame_index_a;
  unsigned db = frame_index_b - frame_index_x;
  for (unsigned i = 0; i<n; ++i)
  {
    //check if the correspondence is present on frame a and on frame b
    if ( !(mask_a[i] && mask_b[i]) )
      continue;
    vnl_vector<double> unknowns(2);
    vgl_point_2d<double> ip = (*camx).project(world_pts_[i]);
    unknowns[0] = ip.x(); unknowns[1] = ip.y();
    double sr = 0, er = 0;
    std::vector<float> cwin = corr_windows_a_[i];
    if (db<da)
      cwin = corr_windows_b_[i];
    this->exhaustive_init(unknowns,win_radius, search_radius,
                          fvx, cwin, sr, er);
    if (verbose_)
      std::cout << "c["<<i <<"]:f[" << frame_index_x << "] Bf(" << ip.x()
               << ' ' << ip.y() << ") Af("  << unknowns[0] << ' '
               << unknowns[1] << ")ier "
               << sr << ':' << er << '\n' << std::flush;
#if 0
    double xb = unknowns[0], yb = unknowns[1];
    if (use_lmq){
      bwm_video_corr_lsqr_cost_func vlcf(fvx, win_radius,
                                         cwin);
      vnl_levenberg_marquardt lmq(vlcf);
      //a good initial guess for the unknown correspondences is
      //the projection of the world points in frame x
      lmq.minimize(unknowns);
      sr = std::sqrt(lmq.get_start_error()/nres);
      er = std::sqrt(lmq.get_end_error()/nres);
    }
    else // amoeba
    {
      bwm_video_corr_cost_function vcf(fvx, win_radius,
                                       cwin);
      vnl_amoeba a(vcf);
      sr = std::sqrt(vcf.f(unknowns)/nres);
      a.minimize(unknowns);
      er = std::sqrt(vcf.f(unknowns)/nres);
    }
    if (verbose_)
      std::cout << "c["<<i <<"]:f[" << frame_index_x << "] Bf(" << xb
               << ' ' << yb << ") Af("  << unknowns[0] << ' '
               << unknowns[1] << ")f "
               << sr << ':' << er << '\n' << std::flush;
#endif
    bwm_video_corr_sptr c = corrs_[i];
    if (er>0)
      c->add(frame_index_x, vgl_point_2d<double>(unknowns[0], unknowns[1]));
    else
      n_failures_[i]++;
  }
  return true;
}

bool bwm_video_corr_processor::
find_missing_correspondences(unsigned win_radius,
                             unsigned search_radius, bool use_lmq)
{
  unsigned n = corrs_.size();
  if (!n){
    std::cerr << "No correspondences\n";
    return false;
  }
  n_failures_.resize(n, 0);
  //get the start and end frame numbers
  unsigned min_frame =0, max_frame = 0;
  std::vector<std::vector<bool> > mask;
  this->mask(min_frame, max_frame, mask);
  //implementation not yet general enough to handle non_zero start frame
  if (min_frame){
    std::cerr << "start frame not zero\n";
    return false;
  }
  unsigned nframes = max_frame - min_frame +1;
  corrs_per_frame_.resize(nframes, 0);
  std::vector<unsigned> frame_intervals;
  for (unsigned f = min_frame; f<=max_frame; ++f)
  {
    //number of correspondences on frame
    unsigned count = 0;
    for (unsigned w = 0; w<n; ++w)
      if (corrs_[w]->match(f))
        count++;
    if (count>=min_corrs)
      frame_intervals.push_back(f);
  }
  unsigned nfi = frame_intervals.size();
  if (nfi<2){
    std::cerr << " not at least two frame bounds\n";
      return false;
  }
  for (unsigned i = 0; i<nfi-1; ++i)
  {
    unsigned frame_a = frame_intervals[i];
    unsigned frame_b = frame_intervals[i+1];
    std::vector<bool> mask_a = mask[frame_a];
    std::vector<bool> mask_b = mask[frame_b];
    for (unsigned ix = frame_a +1; ix<frame_b; ++ix){
      if (!this->find_missing_corrs(frame_a, mask_a,
                                    frame_b, mask_b,
                                    ix, win_radius, search_radius, use_lmq))
        return false;
    }
  }
  for (unsigned ic = 0; ic<n; ++ic)
    std::cout << "nf[" << ic << "]= " << n_failures_[ic]<< '\n';
  for (unsigned f = min_frame; f<=max_frame; ++f)
  {
    for (unsigned ic=0; ic<n; ++ic)
      if (corrs_[ic]->match(f))
        corrs_per_frame_[f]++;
    std::cout << "Nc[" << f << "]= " << corrs_per_frame_[f] << '\n';
  }
  return true;
}

bool bwm_video_corr_processor::refine_world_pts_and_cameras()
{
  //get the start and end frame numbers
  unsigned min_frame =0, max_frame = 0;
  std::vector<std::vector<bool> > mask;
  this->mask(min_frame, max_frame, mask);
  //The implementation is not general enough yet to handle
  //a non-zero start frame
  if (min_frame){
    std::cerr << "Can't handle a non-zero start frame\n";
    return false;
  }
  unsigned ncameras = max_frame-min_frame +1;
  unsigned npoints = corrs_.size();
  if (verbose_)
    std::cout << "Refining " << ncameras << " cameras on " << npoints
             << " correspondences\n";

  //next populate the image points and the mask array
  //the mask appears to be m x n ,
  // where m = number of cameras and n = number of world points (corrs)
  //
  std::vector<vgl_point_2d<double> > image_points(ncameras*npoints);
  for (unsigned w = 0; w<npoints; ++w)
    for (unsigned f = min_frame; f<=max_frame; ++f)
    {
      vgl_point_2d<double> pt;
      if (corrs_[w]->match(f, pt))
        image_points[f*npoints + w] = pt;
      else if (mask[f][w]){
        std::cerr << "fatal[" << f << "][" << w << "]\n";
        return false;
      }
    }
  //find the number of rows that have insufficient correspondences
  //remove them creating a smaller problem
  std::vector<bool> unknown_frames(ncameras, false);
  int filled_rows = 0;
  for (unsigned f = 0; f<ncameras; ++f){
    unsigned nc = 0;
    for (unsigned w = 0; w<npoints; ++w)
      if (mask[f][w]) nc++;
    if (nc<min_corrs)
      unknown_frames[f] = true;
    else
      ++filled_rows;
  }
  if (verbose_)
    std::cout << "Executing bundle adjustment on " << filled_rows
             << " cameras\n";
  //create a new mask array and image point vector for the smaller problem
  std::vector<bool> init(npoints);
  std::vector<std::vector<bool> > cmask(filled_rows, init);
  std::vector<vgl_point_2d<double> > cimage_points;
  std::vector<vpgl_perspective_camera<double> > unknown_cameras;
  if (!cam_istr_)
    return false;
  unsigned ff = 0;
  for (unsigned f = 0; f<ncameras; ++f)
    if (!unknown_frames[f])
    {
      for (unsigned w = 0; w<npoints; ++w)
        if ((cmask[ff][w] = mask[f][w])) // assignment, not a comparison!
          cimage_points.push_back(image_points[f*npoints + w]);

      if (!cam_istr_->seek_camera(f)){
        std::cerr << "Can't seek camera at frame " << f << '\n';
        return false;
      }
      unknown_cameras.push_back(*(cam_istr_->current_camera()));
      ++ff;
    }
  // initialize unknown world points
  vgl_point_3d<double> pun(0.0, 0.0, 0.0);
  std::vector<vgl_point_3d<double> > unknown_world(npoints,pun);
  for (unsigned w = 0; w<npoints; ++w)
    if (corrs_[w]->world_pt_valid())
      unknown_world[w]=corrs_[w]->world_pt();

  // execute the bundle adjustment
  vpgl_bundle_adjust adj;
  /*bool success=*/ adj.optimize(unknown_cameras,
                                 unknown_world, cimage_points,
                                 cmask);
  //save the solved world points in the correspondences
  for (unsigned w = 0; w<npoints; ++w)
  {
    vgl_point_3d<double> pt = unknown_world[w];
    corrs_[w]->set_world_pt(pt);
  }
  std::vector<vpgl_perspective_camera<double> > interp_cameras;
  if (!this->interpolate_cameras(unknown_cameras, unknown_frames,
                                 interp_cameras))
    return false;
  cameras_ = interp_cameras;
  // close the input camera stream
  if (cam_ostr_&&cam_ostr_->is_open())
    for (unsigned i = 0; i<ncameras; ++i)
      cam_ostr_->write_camera(&interp_cameras[i]);
  return true;
}

void bwm_video_corr_processor::
print_frame_alignment_quality(unsigned start_frame, unsigned end_frame)
{
  unsigned n = corrs_.size();
  if (!n) return;
  if (!video_istr_ || !cam_istr_)
    return;

  for (unsigned f = start_frame; f<=end_frame; ++f)
  {
    if (!cam_istr_->seek_camera(f))
      continue;
    vpgl_perspective_camera<double>* cam = cam_istr_->current_camera();
    bsta_histogram<double> h(0.0, 5.0, 10);
    double min_error = vnl_numeric_traits<double>::maxval, max_error = 0;
    for (unsigned i = 0; i<n; ++i)
    {
      bwm_video_corr_sptr c = corrs_[i];
      if (!c->world_pt_valid())
        continue;
      vgl_point_2d<double> cpt;
      if (!c->match(f, cpt))
        continue;
      vgl_point_2d<double> ppt = (*cam).project(c->world_pt());
      double d = vgl_distance(cpt, ppt);
      if (d<min_error) min_error = d;
      if (d>max_error) max_error = d;
      if (d>5.0) h.upcount(5.0, 1.0);
      else h.upcount(d, 1.0);
    }
    std::cout.precision(2);
    std::cout << "frame[" << f << "](" << min_error << '<' << max_error
             << "):c(" << h.p(0.5)+h.p(1.5) << ") p[";
    for (double x = 0.5; x<=5.5; x+=1.0)
      std::cout << h.p(x) << ' ';
    std::cout << "]\n";
  }
}

void bwm_video_corr_processor::close()
{
  if (video_istr_)
    video_istr_->close();
  video_istr_ = nullptr;

  if (cam_istr_)
    cam_istr_->close();
  cam_istr_ = nullptr;

  if (cam_ostr_)
    cam_ostr_->close();
  cam_ostr_ = nullptr;

  site_name_ = "";
  site_path_ = "";
  video_path_ = "";
  camera_path_ = "";

  corrs_.clear();
  cameras_.clear();
  world_pts_.clear();
  corr_windows_a_.clear();
  corr_windows_b_.clear();
  frame_index_a_ = static_cast<unsigned>(-1);
  frame_index_b_ = static_cast<unsigned>(-1);
}
